#define SAVE_SCREEN        // Save/Restore LCD Contents

#include "defs.h"

short _ti89;

/*===========================================================================*/
/* returns unsigned short with key flags set or 0 (TI-89 VERSION) From thomas nussbaumer         */
/*===========================================================================*/
unsigned short GetPossibleKeys89(void) {
    unsigned short key = 0;
    if (_rowread(~(1<<6)) & 0x01) key |= KEYFLAGS_ESC;
    if (_rowread(~(1<<0)) & 0x01) key |= KEYFLAGS_UP;
    if (_rowread(~(1<<0)) & 0x02) key |= KEYFLAGS_LEFT;
    if (_rowread(~(1<<0)) & 0x04) key |= KEYFLAGS_DOWN;
    if (_rowread(~(1<<0)) & 0x08) key |= KEYFLAGS_RIGHT;
    return key;
}

void fidle(void){
	OSSetSR(0x000);	//enable ints
	ngetchx();
	OSSetSR(0x700);	//disable ints
}

void affbackbuf(void *layer){
	 unsigned char* plane2;
    unsigned char* dest1;
    unsigned char* src1;
    int            loop;
    plane2 = LCD_MEM;
    src1  = layer;
	 dest1 = plane2;
	 SetPlane(1);
	 clrscr();
    for (loop=0;loop<100;loop++,src1+=20,dest1+=30) {
        memcpy(dest1,src1,20);
    }
}
/****************************************************************************
** Allocate tables							   **
****************************************************************************/
void GetTables(void){
	cTable = malloc(sizeof(long) * ANGLE_360);
	sTable = malloc(sizeof(long) * ANGLE_360);
	tTable = malloc(sizeof(long) * ANGLE_360);
//	rcTable = malloc(sizeof(long) * ANGLE_360);
//	rsTable = malloc(sizeof(long) * ANGLE_360);
//	rtTable = malloc(sizeof(long) * ANGLE_360);
	nxTable = malloc(sizeof(long) * ANGLE_360);
	nyTable = malloc(sizeof(long) * ANGLE_360);
}
/****************************************************************************
** Check pour un hit dans la  master grid (pendant mouvement)	            **
****************************************************************************/
int CheckHit(long x,long y){
	int	gPosn;
	if (x < 0 || y < 0 || x > GRID_XMAX || y > GRID_YMAX)
		return(0);
	gPosn = ((y >> 6) * GRID_WIDTH) + (x >> 6);
	if (gPosn < 0 || gPosn > GRID_MAX)
		return(0);
	LastPosn = gPosn;
	return(Grid[gPosn]);
}
/****************************************************************************
** Retourne la racine carree d'un long                      				   **
****************************************************************************/
long long_sqrt(long v){
	int				i;
	unsigned	long	result,tmp;
	unsigned	long	low,high;

	if (v <= 1L) return((unsigned)v);
		low = v;
	high = 0L;
	result = 0;
	for (i = 0; i < 16; i++){
		result += result;
		high = (high << 2) | ((low >>30) & 0x3);
		low <<= 2;
		tmp = result + result + 1;
		if (high >= tmp){
			result++;
			high -= tmp;
		}
	}
	return(result);
}

char xRay(long x,long y,int CurAng){
	long		x1 ,y1;
	int		xbeg;
	int		posn;
	long		Adj,nx,ny;
	long		tvalue;
	char		ch;

	if (CurAng == ANGLE_90 || CurAng == ANGLE_270)	/* Don't check vertical lines */
		return(0);
	xbeg = x & 0xFFC0;		/* Same as (X / 64) * 64 to get upper left X */
	tvalue = tTable[CurAng];
	ny = nyTable[CurAng];
	if (!CurAng){
		x1 = xbeg + GRID_XSTEP;
		y1 = 0;
		nx = GRID_XSTEP;
		ny = 0;
	}
	else{
		if (CurAng > ANGLE_270 || CurAng < ANGLE_90){
		x1 = xbeg + GRID_XSTEP;
		Adj = x1 - x;
		y1 = (long)(tvalue * Adj) >> DEC_FACTOR;
		nx = GRID_XSTEP;
		}
	}
	if (CurAng == ANGLE_180){
		x1 = xbeg;
		y1 = 0;
		nx = -GRID_XSTEP;
		ny = 0;
	}
	else{
		if (CurAng > ANGLE_90 && CurAng < ANGLE_270){
			x1 = xbeg;
			Adj = x1 - x;
			y1 = (long)(tvalue * Adj) >> DEC_FACTOR;
			nx = -GRID_XSTEP;
		}
	}
	if (CurAng > ANGLE_180)
		ny = -ny;
	y1 += y;
	while (1){
		if (x1 > GRID_XMAX || y1 > GRID_YMAX ||x1 < 0 || y1 < 0)
			break;
		posn = ((y1 >> 6) * GRID_WIDTH) + (x1 >> 6);
		ch = xGrid[posn];
		if (ch){
			LastX1 = x1;
			LastY1 = y1;
			return(ch);
		}
		x1 += nx;
		y1 += ny;
	}
	return(0);
}

char yRay(long x,long y,int CurAng){
	long    x1,y1;
	int	    ybeg;
	int	    posn;
	long    Opp,nx,ny;
	long    tvalue,rtvalue;
	char    ch;

	if(CurAng == 0 || CurAng == ANGLE_180)	    /* No need to check horiz lines */
		return(0);
	ybeg=y & 0xFFC0;	    /* Same as (Y / 64) * 64 to get upper left Y */
	tvalue=tTable[CurAng];    /* Tangent value for the current angle  */
//	rtvalue=rtTable[CurAng];
	nx=nxTable[CurAng];	    /* X amount for the current angle	    */
	if(CurAng == ANGLE_90){
		x1=0;
		y1=ybeg + GRID_YSTEP;
		ny=GRID_YSTEP;
		nx=0;
	}
	else{
		if(CurAng > 0 && CurAng < ANGLE_180){
			y1=ybeg+GRID_YSTEP;
			Opp=y1-y;
			Opp<<=DEC_FACTOR;
			x1=Opp / tvalue;
			//x1=Opp * rtvalue;
			ny=GRID_YSTEP;
		}
	}
	if (CurAng == ANGLE_270){
		x1=0;
		y1=ybeg;
		ny=-GRID_YSTEP;
		nx=0;
	}
	else{
		if (CurAng > ANGLE_180){
			y1=ybeg;
			Opp=y1-y;
			Opp<<=DEC_FACTOR;
			x1=Opp / tvalue;
			//x1=Opp * rtvalue;
			ny=-GRID_YSTEP;
		}
	}
	if (CurAng > ANGLE_90 && CurAng < ANGLE_270)
		nx=-nx;
	x1+=x;
	while (1){
		if (x1 > GRID_XMAX || y1 > GRID_YMAX ||x1 < 0 || y1 < 0)
			break;
		posn = ((y1 >> 6) * GRID_WIDTH) + (x1 >> 6);
		ch = yGrid[posn];
		if (ch){
			LastX1 = x1;
			LastY1 = y1;
			return(ch);
		}
    x1 += nx;
    y1 += ny;
    }
	return(0);
}

void DrawView(long x,long y,long angle){
	int	CurAng,CalcAng;
	short i,r,ii;
	char	color,xcolor,ycolor;
	long	xDst,yDst,dst,deltax,deltay;
	long	ht,ty,ty1,vx,vy;

	r=80;
	ii=res+1;
	memset(vscreen,0,2000);
	PortSet(vscreen,159,99);
	CurAng = angle+85;	//171 pour resolution de 160
	for (i = 0; i < r; i=i+ii){
		if (CurAng < 0)
			CurAng = ANGLE_360 + CurAng;
		if (CurAng >= ANGLE_360)
			CurAng = CurAng - ANGLE_360;
		CalcAng = abs(angle - CurAng);
		if (CalcAng < 0)
			CalcAng = ANGLE_360 + CalcAng;
		if (CalcAng >= ANGLE_360)
			CalcAng = CalcAng - ANGLE_360;
		xDst = yDst = 500000L;	    /* Set to wild distance for later compare */
		xcolor = xRay(x,y,CurAng);
		/* A hit on an X wall */
		if (xcolor!=0){
			deltax = labs(LastX1 - x);
			deltay = labs(LastY1 - y);
			xDst = long_sqrt((deltax * deltax) + (deltay * deltay));
		}
		ycolor = yRay(x,y,CurAng);
		/* A hit on a Y wall */
		if (ycolor!=0){
			deltax = labs(LastX1 - x);
			deltay = labs(LastY1 - y);
			yDst = long_sqrt((deltax * deltax) + (deltay * deltay));
		}
		if (xcolor || ycolor){
			if (xDst <= yDst){
				color = xcolor;
				dst = xDst;
			}
			else{
				color = ycolor;
				dst = yDst;
			}
			vx = cTable[CalcAng];
			if (vx){
				vy = dst * vx;
				dst = vy >> DEC_FACTOR;
			}
			if (dst > 3071)
				dst = 3071;
			ht = DstTable[dst];
			ty = 50 - (ht>>1);
			ty1 = ty + ht;
			if (ty < 0) ty = 0;
				if (ty1 > 99) ty1 = 99;
		}
		if(res==0){
			DrawLine(i,(short)ty,i,(short)ty1,A_NORMAL);
		}else{
			DrawLine(i,(short)ty,i,(short)ty1,A_NORMAL);
			DrawLine(i+1,(short)ty,i+1,(short)ty1,A_NORMAL);
		}
		CurAng=CurAng-ii;
	}
	affbackbuf(vscreen);
}
void loadsystem(void){
	int	    done;
	long   x,y,x1,y1;
	float p=0.0;
	float i360=1.0/1024;
	float   ang,ca;

	vscreen=malloc(2000);
	FontSetSys(F_4x6);
	OSSetSR(0x0700);
	GetTables();
	y1 = 12000L;
	DstTable[0] = y1;
	for (done = 1; done < 1024; done += 1){
		x1 = y1 / done;
		if (y1 - (x1 * done) > (done / 2))
		x1++;
		DstTable[done] = x1;
	}
	ang = 0;
	ca = 6.283185307 / ANGLE_360;
	clrscr();
	printf_xy(0,0,"      _TIRayX Engine_ V0.10 Coded By Alkher");
	printf("\n\nLoading trig tables");
	for (x = 0; x < ANGLE_360; x++){
		printf_xy(0,18,"Progress: _%.1f%%     %l",100*p*i360,tTable[x]);
		p++;
		tTable[x] = (long)(tan(ang) * DEC_MULT);
		cTable[x] = (long)(cos(ang) * DEC_MULT);
		sTable[x] = (long)(sin(ang) * DEC_MULT);
//		rtTable[x] = (long)(1/(tan(ang) * DEC_MULT));
//		rcTable[x] = (long)(1/(cos(ang) * DEC_MULT));
//		rsTable[x] = (long)(1/(sin(ang) * DEC_MULT));
		nxTable[x] = 64L;
		if (tTable[x])
			nxTable[x] = labs((64L << DEC_FACTOR) / tTable[x]);
		y = 64L * tTable[x];
		nyTable[x] = labs(y >> DEC_FACTOR);
		ang += ca;
	}
	printf("     _Done");
}

void shutdown(void){
	OSSetSR(0x0000);
	PortRestore();
	free(vscreen);
	free(cTable);
	free(sTable);
	free(tTable);
//	free(rcTable);
//	free(rsTable);
//	free(rtTable);
	free(nxTable);
	free(nyTable);
}

void _main(){
	int	    i,done;
	long   x,y,x1,y1,angle;
	unsigned short key;
	unsigned long ram;
	
	loadsystem();
	HeapCompress(); // compress heap first !!!    
	ram=HeapAvail();
	printf_xy(0,24,"_Free RAM Detected: %lu",ram);
	printf_xy(0,36,"Press ENTER to continue");
	fidle();
	y1 = (GRID_WIDTH+1) * (GRID_HEIGHT+1);
	memset(xGrid,0,y1);
	memset(yGrid,0,y1);
	for (y = 0; y < GRID_HEIGHT; y++){
		for (x = 0; x < GRID_WIDTH; x++){
			y1 = (y * GRID_WIDTH) + x;
			done = Grid[(y * GRID_WIDTH) + x];
			if (done){
				xGrid[y1] = done;
				xGrid[y1+1] = done;
				yGrid[y1] = done;
				yGrid[y1+GRID_WIDTH] = done;
			}
		}
	}
	x = 96;
	y = 96;
	angle = 0;
	done = 0;
	DrawView(x,y,angle);
	i=0;
	while (!done){
		if (key = GetPossibleKeys89()){
			if (key & KEYFLAGS_ESC)   break;
			if (key & KEYFLAGS_UP){
				x1 = x +2+2* ((cTable[angle] * 2)) / DEC_MULT;
				y1 = y +2+2* ((sTable[angle] * 2)) / DEC_MULT;
				if (!CheckHit(x1,y1)){
					x = x1;
					y = y1;
				}
			}
			if (key & KEYFLAGS_DOWN){
				x1 = x -2-2* ((cTable[angle] * 2)) / DEC_MULT;
				y1 = y -2-2* ((sTable[angle] * 2)) / DEC_MULT;
				if (!CheckHit(x1,y1)){
					x = x1;
					y = y1;
				}
			}
			if (key & KEYFLAGS_RIGHT){
				angle -= 8;
				if (angle < 0)
					angle = ANGLE_360 + angle;
			}
			if (key & KEYFLAGS_LEFT){
				angle += 8;
				if (angle >= ANGLE_360)
					angle = angle - ANGLE_360;
			}
		}
		i++;
		DrawView(x,y,angle);
	}
}
